print("import: 'eelbrain'")
import eelbrain

print("import: 'eelbrain.load'")
import eelbrain.load

print("import: 'eelbrain.mne_fixes'")
import eelbrain.mne_fixes

print("import: 'eelbrain.plot'")
import eelbrain.plot

print("import: 'eelbrain.save'")
import eelbrain.save

